"""Layer to perform attentive scoring."""

import math

from lingvo import compat as tf
from lingvo.core import base_layer
from lingvo.core import py_utils


class AttentiveScoringLayer(base_layer.BaseLayer):
  """Calculates attentive scores for enroll & test set representations."""

  @classmethod
  def Params(cls):
    p = super(AttentiveScoringLayer, cls).Params()
    p.name = 'attentive_scoring_layer'
    p.Define(
        'num_keys', 32, 'How many keys to use for attentive scoring. Must '
        'select one or more. (Default: 32)')
    p.Define(
        'key_dim', 16,
        'The number of elements to use for each key vector. Must use 1 or more.'
        ' (Default: 16)')
    p.Define(
        'value_dim', 48,
        'The number of elements to use for each value vector. Must use 1 or '
        'more. (Default: 48)')
    p.Define(
        'use_keys_and_queries', False, 'If set to True, uses separate keys'
        ' and queries in the standard way that attention is applied to '
        'perform the comparison (i.e. keys and queries are estimated '
        'independently). If False, keys are treated as both keys and queries '
        '(i.e. they are the same). (Default: False)')
    p.Define(
        'scale_factor', 16.0, 'The value to multiply the key scores by to '
        'determine how aggressive the softmax function should be. '
        '(Default: 16.0)')
    p.Define(
        'use_trainable_scale_factor', True, 'Whether or not to use a trainable'
        ' scaling factor for the key score softmax calculation. This overrides '
        'the static scale_factor setting if set to True. Additionally, the '
        'scale_factor setting will be used to initialize the equivalent '
        'parameter for training purposes. (Default: True)')
    p.Define(
        'apply_l2_norm_to_keys', True, 'Should L2 length normalization be '
        'applied to the key vectors? (Default: True)')
    p.Define(
        'apply_l2_norm_to_values', False, 'Should L2 length normalization be '
        'applied to the value vectors? (Default: False)')
    p.Define(
        'apply_global_l2_norm_to_concat_form', True, 'Should L2 length '
        'normalization be applied to the (global) concatenated attention vector'
        ' result? This normalization is applied after the application of L2 '
        'normalization to keys, queries and values and after the calculation of'
        ' the attention weights. The attention formulation that is represented '
        'as a weighted combination of enrollment and test value vectors can '
        'also be represented as a single high dimensional dot product. This '
        'normalization applies L2-norm to the high dimensional vectors in the '
        'dot product formulation. (Default: True)')
    p.Define(
        'apply_softmax_per_test_key', False, 'Instead of applying softmax '
        'across all key comparisons, apply softmax for each test utterance '
        'key across all other enrollment keys. (Default: False)')

    return p

  def __init__(self, params):
    super().__init__(params)
    p = self.params

    # Check that the parameters are valid.
    if p.num_keys < 1:
      raise ValueError('The parameter num_keys must be an integer greater than '
                       '0.')
    if p.key_dim < 1:
      raise ValueError('The parameter key_dim must be an integer greater than '
                       '0.')
    if p.value_dim < 1:
      raise ValueError('The parameter value_dim must be an integer greater than'
                       ' 0.')
    if p.scale_factor <= 0:
      raise ValueError('The parameter scale_factor must be a real value greater'
                       ' than 0.')

    if p.use_trainable_scale_factor:
      weight_config = py_utils.WeightParams(
          shape=[],
          init=py_utils.WeightInit.Constant(scale=math.log(p.scale_factor)),
          dtype=p.dtype,
          collections=[self.__class__.__name__ + '_vars'])
      self.CreateVariable('trainable_log_scale_factor', weight_config)

  def FProp(self, inputs, theta=None):
    """Performs attentive scoring given enrollment and test representations.

    Args:
      inputs: A tuple of 2 items. The first is the test representations and the
        second is the enrollment speaker representations.
        Test tensors are tf.float32 tensors of size:
        [num_test_utts, representation_dim].
        Enrollment tensors are tf.float32 tensors of size:
        [num_enroll_spks, num_enroll_utts_per_spk, representation_dim].
        Note here that the representation_dim is the total number of attributes
        to describe a single utterance. For example, for 2 keys, where there is
        a key dimensionality of 4 and a value dimensionality of 10, the
        representation_dim would be 2 * (4 + 10) = 28. The representation_dim
        contains all statistics related to a single utterance.

        Each utterance representation is composed of the concatenated result
        of num_keys sets of statistics. The first key_dim numbers of each
        set contains the key and the next value_dim numbers holds the
        corresponding values. The total dimensionality of each utterance
        representation may be calculated as num_keys * (key_dim + value_dim).

        Note that the dimensionality for both test utterance and
        enrollment utterance representations is the same.
      theta: A NestedMap containing layer weights specific to the
        AttentiveScoringLayer layer. Can be set to None (default) if a trainable
        scaling factor is not needed. Specifically, if
        use_trainable_scale_factor is set, theta must contain the trainable
        variable referenced as theta.trainable_log_scale_factor.
    Returns:
      scores: The output scores comparing all test utterance representations
        with all enrollment speaker models (comprising multiple utterances).
        The result is a tf.float32 tensor of size:
        [num_test_utts, num_enroll_spks(or num_models)]

    """

    p = self.params

    # Retrieve the test and enrollment data
    (test_data, enroll_data) = inputs

    # Determine the input parameters
    representation_dim = enroll_data.shape[2]
    test_representation_dim = test_data.shape[1]

    # Check that the sizes are valid
    if representation_dim != test_representation_dim:
      raise ValueError('The enrollment representation_dim (dim=%d) must be the '
                       'same as the test_representation_dim (dim=%d).' %
                       (representation_dim, test_representation_dim))

    if p.use_keys_and_queries:
      expected_representation_dim = p.num_keys * (2 * p.key_dim + p.value_dim)
      if expected_representation_dim != representation_dim:
        raise ValueError(
            'The expected_representation_dim (dim=%d) must be the same as the '
            'actual representation_dim (dim=%d). The '
            'expected_representation_dim is calculated as num_keys * (2 * '
            'key_dim + value_dim) or %d * (2 * %d + %d).' %
            (expected_representation_dim, representation_dim, p.num_keys,
             p.key_dim, p.value_dim))
    else:
      expected_representation_dim = p.num_keys * (p.key_dim + p.value_dim)
      if expected_representation_dim != representation_dim:
        raise ValueError(
            'The expected_representation_dim (dim=%d) must be the same as the '
            'actual representation_dim (dim=%d). The '
            'expected_representation_dim is calculated as num_keys * (key_dim +'
            ' value_dim) or %d * (%d + %d).' %
            (expected_representation_dim, representation_dim, p.num_keys,
             p.key_dim, p.value_dim))

    # Extract keys, queries and values from the utterance representations and
    # apply L2 normalization on each if enabled. If p.use_keys_and_queries is
    # True, keys and queries will be set the same.
    # Note: all tensors are composed of elements of type tf.float32.
    #
    # For enrollment data:
    # enroll_data_reshaped has size:
    # [num_enroll_spks, num_enroll_utts_per_spk, representation_dim]
    # enroll_keys has size:
    # [num_enroll_spks, num_enroll_utts_per_spk, num_keys, key_dim]
    # enroll_values has size:
    # [num_enroll_spks, num_enroll_utts_per_spk, num_keys, value_dim]
    #
    # For test data:
    # test_data has size: [num_test_utts, representation_dim]
    # test_keys has size: [num_test_utts, num_keys, key_dim]
    # test_values has size: [num_test_utts, num_keys, value_dim]
    (enroll_keys, _, enroll_values) = (self._extract_and_norm(enroll_data))
    (_, test_queries, test_values) = self._extract_and_norm(test_data)

    # Determine whether to have a static or trainable scale factor
    if p.use_trainable_scale_factor:
      # Note that exp(log_scale_factor) resolves to the regular
      # scale_factor parameter. This is to ensure that values are positive.
      if hasattr(theta, 'trainable_log_scale_factor'):
        scale_factor = tf.math.exp(theta.trainable_log_scale_factor)
      else:
        raise ValueError(
            'The trainable weight was not specified as part of the forward '
            'propagation function. A trainable scale factor was requested '
            '(i.e. use_trainable_scale_factor was set to True) without '
            'specifying theta.trainable_log_scale_factor.')
    else:
      scale_factor = p.scale_factor

    # Generate all cross scores for keys and values.
    # For the key scores, scale by scale_factor so that it is ready for the
    # softmax calculation. For both keys and values, the output shape is:
    # [num_test_utts, num_keys(test), num_enroll_spks, num_enroll_utts_per_spk,
    # num_keys(enroll)]
    keys_cross_scores_scaled = scale_factor * tf.tensordot(
        test_queries, enroll_keys, axes=((2), (3)))
    values_cross_scores = tf.tensordot(
        test_values, enroll_values, axes=((2), (3)))

    if p.apply_softmax_per_test_key:
      # Calculate the softmax across scores specific to each test key for a
      # trial. Also scale down by the number of test keys so that the final
      # summation is correct.
      scoring_softmax = tf.keras.activations.softmax(
          keys_cross_scores_scaled, axis=[3, 4]) / p.num_keys
    else:
      # Calculate softmax across all key scores specific to each trial
      scoring_softmax = tf.keras.activations.softmax(
          keys_cross_scores_scaled, axis=[1, 3, 4])

    # Calculate scores of test segments against bundled enrollment segments
    # normalized by the softmax weighting function.
    scores = tf.math.reduce_sum(
        tf.multiply(values_cross_scores, scoring_softmax), axis=(1, 3, 4))

    # Check if we will be applying an additional L2-norm for the effective
    # concatenated attention vector
    if p.apply_global_l2_norm_to_concat_form:
      concat_norm = self._calc_l2_norm_for_concat_form(scoring_softmax,
                                                       test_values,
                                                       enroll_values)
      scores = tf.math.divide(scores, concat_norm)

    return scores

  def _extract_and_norm(self, representation_data):
    """Extracts keys and values and L2 normalizes them if requested.

    Args:
      representation_data: The fixed dimensional representations for each
        utterance. Tensors can be of two different shapes. For enrollment data
        they are tf.float32 (3d) tensors of size:
        [num_enroll_spks, num_enroll_utts_per_spk, representation_dim].
        For test data, they are tf.float32 (2d) tensors of size:
        [num_test_utts, representation_dim].
        For both cases, the parameter representation_dim is equivalent to
        num_keys * (key_dim + value_dim).

    Returns:
      keys: The extracted keys. Tensors can be of two different shapes. For
        enrollment data they are tf.float32 (4d) tensors of size:
        [num_enroll_spks, num_enroll_utts_per_spk, num_keys, key_dim]
        For test data they are tf.float32 (3d) tensors of size:
        [num_test_utts, num_keys, key_dim]
        These may be L2-normalized along the key_dim axis if the
        apply_l2_norm_to_keys attribute is True and unnormalized if False.
      queries: The extracted queries. If p.use_keys_and_queries is set to True,
        then the queries will be independent of the keys. If
        p.use_keys_and_queries is False, then the queries returned will be the
        same as the keys. Tensors can be of two different shapes.
        For enrollment data they are tf.float32 (4d) tensors of size:
        [num_enroll_spks, num_enroll_utts_per_spk, num_keys, key_dim]
        For test data they are tf.float32 (3d) tensors of size:
        [num_test_utts, num_keys, key_dim]
        These may be L2-normalized along the key_dim axis if the
        apply_l2_norm_to_keys attribute is True and unnormalized if False.
      values: The extracted values. Tensors can be of two different shapes. For
        enrollment data they are tf.float32 (4d) tensors of size:
        [num_enroll_spks, num_enroll_utts_per_spk, num_keys, value_dim]
        For test data they are tf.float32 (3d) tensors of size:
        [num_test_utts, num_keys, value_dim]
        These may be L2-normalized along the value_dim axis if the
        apply_l2_norm_to_values attribute is True and unnormalized if False.
    """

    p = self.params

    if p.use_keys_and_queries:
      key_query_value_dim = 2 * p.key_dim + p.value_dim
      query_start_ndx = p.key_dim
      query_stop_ndx = 2 * p.key_dim
    else:
      key_query_value_dim = p.key_dim + p.value_dim
      query_start_ndx = 0
      query_stop_ndx = p.key_dim

    # Extract the key and value tensors specific to whether the data is test
    # data (2 axes) or enrollment data (3 axes).
    data_shape = representation_data.shape.as_list()
    if len(data_shape) == 2:
      # Test data with shape [num_test_utts, representation_dim]
      # converts to keys with shape [num_test_utts, num_keys, key_dim]
      # and values with shape [num_test_utts, num_keys, value_dim]
      num_test_utts = data_shape[0]
      data_reshaped = tf.reshape(
          representation_data,
          shape=[num_test_utts, p.num_keys, key_query_value_dim])
      keys = data_reshaped[:, :, 0:p.key_dim]
      queries = data_reshaped[:, :, query_start_ndx:query_stop_ndx]
      values = data_reshaped[:, :, query_stop_ndx:]
    else:
      # Enrollment data with shape [num_enroll_spks, num_enroll_utts_per_spk,
      # representation_dim] converts to keys with shape [num_enroll_spks,
      # num_enroll_utts_per_spk, num_keys, key_dim] and values with shape
      # [num_enroll_spks, num_enroll_utts_per_spk, num_keys, value_dim]
      (num_enroll_spks, num_enroll_utts_per_spk) = data_shape[0:2]
      data_reshaped = tf.reshape(
          representation_data,
          shape=[
              num_enroll_spks, num_enroll_utts_per_spk, p.num_keys,
              key_query_value_dim
          ])
      keys = data_reshaped[:, :, :, 0:p.key_dim]
      queries = data_reshaped[:, :, :, query_start_ndx:query_stop_ndx]
      values = data_reshaped[:, :, :, query_stop_ndx:]

    # If enabled, perform unit length normalization on keys (or queries)
    if p.apply_l2_norm_to_keys:
      keys = tf.nn.l2_normalize(keys, axis=-1)
      queries = tf.nn.l2_normalize(queries, axis=-1)

    # If enabled, perform unit length normalization on values
    if p.apply_l2_norm_to_values:
      values = tf.nn.l2_normalize(values, axis=-1)

    return (keys, queries, values)

  def _calc_l2_norm_for_concat_form(self, scoring_softmax, test_values,
                                    enroll_values):
    """Calculates the L2-norm for the effective concatenated attention vector.

    Args:
      scoring_softmax: The softmax weightings to apply to the (trial based)
        cross scores. Tensors are of tf.float32 with size: [num_test_utts,
          num_keys(test), num_enroll_spks, num_enroll_utts_per_spk,
          num_keys(enroll)].
      test_values: The test vector related values (tf.float32) of size:
        [num_test_utts, num_keys, value_dim]
      enroll_values: The enrollment vector related values (tf.float32) of size:
        [num_enroll_spks, num_enroll_utts_per_spk, num_keys, value_dim]

    Returns:
      l2_norm_concat: The effective L2-norm to divide the attention scores by.
        The result is tf.float32 of size: [num_test_utts, num_enroll_speakers].
    """

    # Test embeddings
    # Result is 2d: [num_test_utts, num_keys]
    test_values_sumsq = tf.math.reduce_sum(
        tf.math.multiply(test_values, test_values), axis=2)

    # Result is 2d: [num_test_utts, num_enroll_speakers]
    test_values_sumsq_weighted = tf.math.sqrt(
        tf.reduce_sum(
            tf.math.multiply(
                scoring_softmax, test_values_sumsq[:, :, tf.newaxis, tf.newaxis,
                                                   tf.newaxis]),
            axis=(1, 3, 4)))

    # Enroll embeddings
    # Result is 3d: [num_enroll_spks, num_enroll_utts_per_spk, num_keys]
    enroll_values_sumsq = tf.math.reduce_sum(
        tf.math.multiply(enroll_values, enroll_values), axis=3)

    # Result is 2d: [num_test_utts, num_enroll_speakers]
    enroll_values_sumsq_weighted = tf.math.sqrt(
        tf.reduce_sum(
            tf.math.multiply(
                scoring_softmax, enroll_values_sumsq[tf.newaxis,
                                                     tf.newaxis, :, :, :]),
            axis=(1, 3, 4)))

    # Determine the matrix to normalize the cross scores by
    # Result is 2d: [num_test_utts, num_enroll_speakers]
    l2_norm_concat = tf.math.multiply(test_values_sumsq_weighted,
                                      enroll_values_sumsq_weighted)

    return l2_norm_concat



"""Utilities for the audio feature frontend."""

import math
import librosa
from lingvo.tasks.asr import frontend as lingvo_frontend
import numpy as np
from scipy.io import wavfile
import tensorflow.compat.v2 as tf


def read_mean_stddev_csv(csv_file: str) -> tuple[np.ndarray, np.ndarray]:
  """Reads mean and stddev from a CSV file."""
  with open(csv_file) as f:
    csv_data = np.genfromtxt(f, delimiter=",")
  return csv_data[0, :], csv_data[1, :]


def get_int_samples(file_name: str) -> np.ndarray:
  with open(file_name, "rb") as f:
    sample_rate, int_samples = wavfile.read(f)

  if sample_rate != 16000:
    samples = int_samples / 32768.0
    samples = librosa.resample(samples, sample_rate, target_sr=16000)
    int_samples = (samples * 32768.0).astype(np.int16)

  return np.expand_dims(int_samples, axis=0)


def get_frontend_layer() -> lingvo_frontend.MelAsrFrontend:
  """Get frontend layer."""
  fe_params = {
      "frame_size_ms": 32,
      "num_bins": 128,
      "sample_rate": 16000.0,
      "lower_edge_hertz": 125,
      "upper_edge_hertz": 7500,
      "preemph": 0.97,
      "noise_scale": 8,
      "pad_end": False,
      "fft_overdrive": True,
      "compute_energy": False,
      "stack_left_context": 3,
      "stack_right_context": 0,
      "frame_stride": 3,
      "per_bin_mean": None,
      "per_bin_stddev": None,
  }

  p = lingvo_frontend.MelAsrFrontend.Params().Set(**fe_params)
  return p.Instantiate()


def add_cluster_id(
    features: np.ndarray, cluster_id: int, num_clusters: int
) -> np.ndarray:
  if not num_clusters:
    return features
  cluster_ids = np.zeros((features.shape[0], num_clusters))
  cluster_ids[:, cluster_id] = 1
  return np.concatenate([features, cluster_ids], axis=1).astype(np.float32)


def normalize_features(
    features: np.ndarray, mean_stddev: tuple[np.ndarray, np.ndarray]
) -> np.ndarray:
  vad_mean, vad_std_dev = mean_stddev
  mean_vec = np.expand_dims(np.repeat(vad_mean, 4), axis=0)
  std_dev_vec = np.expand_dims(np.repeat(vad_std_dev, 4), axis=0)
  features = (features - mean_vec) / std_dev_vec
  return features.astype(np.float32)


def load_tflite_model(model_path: str) -> tf.lite.Interpreter:
  """Reads a serialized TFLite model and returns the Interpreter object."""
  # Read model.
  with open(model_path) as file_object:
    model_content = file_object.read()
  tflite_model = tf.lite.Interpreter(model_content=model_content)
  tflite_model.allocate_tensors()
  tflite_model.reset_all_variables()
  return tflite_model


def run_single_input_model_one_step(
    model: tf.lite.Interpreter, input_data: np.ndarray
) -> np.ndarray:
  """Runs given TFLite model (as interpreter) on data for a single step."""
  model.set_tensor(
      model.get_input_details()[0]["index"],
      input_data,
  )
  model.invoke()
  return model.get_tensor(model.get_output_details()[0]["index"])


def print_model_info(model: tf.lite.Interpreter) -> None:
  """Returns input and output tensor details."""
  model.reset_all_variables()
  input_details = model.get_input_details()
  output_details = model.get_output_details()
  print("input_details:", input_details)
  print("output_details:", output_details)


def apply_vad(
    features: np.ndarray,
    vad_model: tf.lite.Interpreter,
    mean_stddev: tuple[np.ndarray, np.ndarray],
    threshold: float,
    cluster_id: int = 2,
    num_clusters: int = 16,
) -> tuple[np.ndarray, np.ndarray]:
  """Applies VAD to the features."""
  normalized_features = normalize_features(features, mean_stddev)
  normalized_features = add_cluster_id(
      normalized_features, cluster_id, num_clusters
  )

  vad_decisions = []
  for t in range(normalized_features.shape[0]):
    features_t = normalized_features[t, :]
    features_t = np.expand_dims(features_t, axis=0)
    output = run_single_input_model_one_step(vad_model, features_t)
    vad_score = math.exp(output[0, 0])
    vad_decisions.append(vad_score > threshold)
  vad_decisions = np.array(vad_decisions)

  features_after_vad = features[vad_decisions, :]
  return vad_decisions, features_after_vad


def run_multi_input_model_one_step(
    model: tf.lite.Interpreter, input_data: np.ndarray, states: list[np.ndarray]
) -> tuple[np.ndarray, list[np.ndarray]]:
  """Runs given TFLite model (as interpreter) on data for a single step."""
  model.set_tensor(
      model.get_input_details()[0]["index"],
      input_data,
  )
  for i, state in enumerate(states):
    model.set_tensor(
        model.get_input_details()[i + 1]["index"],
        state,
    )
  model.invoke()
  output_data = model.get_tensor(model.get_output_details()[0]["index"])
  output_states = []
  for i in range(len(states)):
    output_states.append(
        model.get_tensor(model.get_output_details()[i + 1]["index"])
    )
  return output_data, output_states


def run_multi_input_model(
    model: tf.lite.Interpreter, features: np.ndarray, batch_size: int = 2
) -> np.ndarray:
  """Runs given TFLite model (as interpreter) on data for a sequence."""
  states = []
  outputs = []
  for i in range(len(model.get_input_details()) - 1):
    states.append(
        np.zeros(
            shape=model.get_input_details()[i + 1]["shape"],
            dtype=model.get_input_details()[i + 1]["dtype"],
        )
    )
  for t in range(int(features.shape[0] / batch_size)):
    start_index = t * batch_size
    features_t = features[start_index: start_index + batch_size, :]
    output, states = run_multi_input_model_one_step(model, features_t, states)
    outputs.append(output)
  return np.concatenate(outputs, axis=0)



"""Tests for utils."""

from lingvo import compat as tf
from lingvo.core import test_utils

import numpy as np

import utils

_RANDOM_SEED = 1321


class SVLUtilsTest(test_utils.TestCase):

  def setUp(self):
    super().setUp()

    self.seq_len = 5
    self.feature_dim = 2
    self.num_spks_per_batch = 2
    self.num_utts_per_spk = 2

  def testGetLastSeqOutput(self):
    with self.session() as sess:
      seq = tf.constant(
          list(
              range(self.seq_len * self.num_spks_per_batch *
                    self.num_utts_per_spk * self.feature_dim)))
      seq = tf.reshape(seq, [
          self.num_spks_per_batch * self.num_utts_per_spk, self.seq_len,
          self.feature_dim
      ])
      seq = tf.transpose(seq, [1, 0, 2])
      paddings = []
      lengths = [2, 1, 5, 3]
      for i in lengths:
        assert i <= self.seq_len
        example = tf.zeros([1, i], dtype=tf.float32)
        padded_example = tf.pad(
            example, [[0, 0], [0, self.seq_len - i]], constant_values=1.0)
        print(padded_example)
        paddings.append(padded_example)
      padding = tf.concat(paddings, axis=0)
      padding = tf.transpose(padding, [1, 0])
      last_seq_output = sess.run(utils.GetLastSeqOutput(seq, padding))
      expected_output = np.asarray([[2, 3], [10, 11], [28, 29], [34, 35]])
      np.testing.assert_array_almost_equal(
          last_seq_output, expected_output, decimal=4)

  def testComputeSimilaritySimple(self):
    with self.session() as sess:
      logits = tf.constant([[1.0, 2.0], [1.0, 5.0], [-2.0, -1.0], [-5.0, -1.0]],
                           tf.float32)
      scores = sess.run(
          utils.ComputeSimilaritySimple(logits, self.num_spks_per_batch,
                                        self.num_utts_per_spk))
      # Expect the same speaker utt scores to the centroid being the same,
      # because it is symmetric when there are only 2 utts per speaker.
      expected_scores = np.array([[0.991152, -0.713282], [0.991152, -0.503735],
                                  [-0.713282, 0.991152], [-0.503735, 0.991152]],
                                 np.float32)
      np.testing.assert_array_almost_equal(scores, expected_scores, decimal=4)

  def testComputeSimilaritySplit(self):
    with self.session() as sess:
      logits = tf.constant([[1.0, 2.0], [1.0, 5.0], [-2.0, -1.0], [-5.0, -1.0],
                            [2.0, 2.0], [3.0, 5.0], [-4.0, -1.0], [-6.0, -1.0]],
                           tf.float32)
      self.num_utts_per_spk = 4
      tf.random.set_seed(_RANDOM_SEED)

      scores = sess.run(
          utils.ComputeSimilaritySplit(logits, self.num_spks_per_batch,
                                       self.num_utts_per_spk))

      # Expect positive trial speaker scores to be lower than the scores from
      # ComputeSimilaritySimple, because for each utt, it is not taken for its
      # own speaker centroid computation, AKA there is no cheating.
      expected_scores = np.array(
          [[0.3162278, 0.4876413], [0.55470014, 0.7566749],
           [0.3162278, 0.13371336], [0.55470014, 0.31247067], [0., 0.1865364],
           [0.24253565, 0.4926988], [0.5144958, 0.34551173],
           [0.58123815, 0.34293512]], np.float32)

      np.testing.assert_array_almost_equal(scores, expected_scores, decimal=4)

  def testComputeSimilaritySplitWithVaryingEnroll(self):
    with self.session() as sess:
      logits = tf.constant([[1.0, 2.0], [1.0, 5.0], [-2.0, -1.0], [-5.0, -1.0],
                            [2.0, 2.0], [3.0, 5.0], [-4.0, -1.0], [-6.0, -1.0]],
                           tf.float32)
      self.num_utts_per_spk = 4
      tf.random.set_seed(_RANDOM_SEED)

      # Vary the number of enrollments (last parameter set to True)
      scores = sess.run(
          utils.ComputeSimilaritySplit(logits, self.num_spks_per_batch,
                                       self.num_utts_per_spk, True))

      expected_scores = np.array(
          [[0.3162278, 0.9970546], [-0.6139406, 0.7566749],
           [0.3162278, -0.84366155], [0.9647638, 0.31247067], [0., 0.97014254],
           [-0.84366167, 0.4926988], [0.5144958, -0.7071068],
           [0.95577914, 0.34293512]], np.float32)

      np.testing.assert_array_almost_equal(scores, expected_scores, decimal=4)


if __name__ == '__main__':
  tf.test.main()



"""A library to run lang-id TFLite model inference."""

import dataclasses

from lingvo.core import py_utils
import numpy as np

import fe_utils
import language_map


@dataclasses.dataclass
class WavToLangRunner:
  """WavToLangRunner."""

  # Path to the VAD TFLite model.
  vad_model_file: str = ""
  # Path to the VAD mean and stddev CSV file.
  vad_mean_stddev_file: str = ""
  # Path to the lang-id TFLite model.
  langid_model_file: str = ""

  vad_threshold: float = 0.1
  vad_cluster_id: int = 2
  vad_num_clusters: int = 16

  def __post_init__(self):
    self.vad_model = fe_utils.load_tflite_model(self.vad_model_file)
    self.langid_model = fe_utils.load_tflite_model(self.langid_model_file)
    self.vad_mean_stddev = fe_utils.read_mean_stddev_csv(
        self.vad_mean_stddev_file
    )

  def wav_to_lang(self, audio_file: str) -> tuple[str, np.ndarray]:
    """Run lang-id model on wav file."""
    input_signal = fe_utils.get_int_samples(audio_file)
    return self.samples_to_lang(input_signal)

  def samples_to_lang(self, input_signal: np.ndarray) -> tuple[str, np.ndarray]:
    """Run lang-id model on int16 samples."""
    input_paddings = fe_utils.np.zeros(input_signal.shape[:2])

    fe = fe_utils.get_frontend_layer()
    result = fe.FProp(
        fe.theta,
        py_utils.NestedMap(src_inputs=input_signal, paddings=input_paddings),
    )
    features = result.src_inputs[0, :, :, 0].numpy()

    self.vad_model.reset_all_variables()
    _, features_after_vad = fe_utils.apply_vad(
        features,
        self.vad_model,
        self.vad_mean_stddev,
        self.vad_threshold,
        self.vad_cluster_id,
        self.vad_num_clusters,
    )

    self.langid_model.reset_all_variables()
    langid_outputs = fe_utils.run_multi_input_model(
        self.langid_model, features_after_vad
    )

    last_langid_frame = langid_outputs[-1, :]
    language_code = language_map.ID_TO_LANG[int(np.argmax(last_langid_frame))]

    return language_code, last_langid_frame



"""Tests for cumulative_statistics_layer."""

from absl.testing import parameterized
from lingvo import compat as tf
from lingvo.core import py_utils
from lingvo.core import test_utils
import numpy as np

import cumulative_statistics_layer

_TF_RANDOM_SEED = 1314


class CumulativeStatisticsLayerTest(test_utils.TestCase,
                                    parameterized.TestCase):
  """The cumulative statistics layer tests."""

  def setUp(self):
    """Generates test minibatches, the expected output and the layer config."""

    super().setUp()

    tf.reset_default_graph()

    # Generate the common list of minibatches
    self.input_batch_list = self._GenerateListOfMinibatches()

    # Get the expected test results to compare with
    self.expected_outputs = self._GetExpectedTestOutputs()

    # Set up the basic cumulative statistics layer config
    input_dim = self.input_batch_list[0].features.shape[2]
    p = cumulative_statistics_layer.CumulativeStatisticsLayer.Params()
    p.stats_type = 'PASS_THRU'
    p.use_weighted_frames = False
    p.input_dim = input_dim
    p.features_name = 'features'
    p.paddings_name = 'paddings'
    p.frame_weight_ffn.activation = ['SIGMOID']
    p.frame_weight_ffn.has_bias = [True]
    p.frame_weight_ffn.hidden_layer_dims = [1]
    p.frame_weight_ffn.input_dim = input_dim
    self.params = p

  def _GenerateListOfMinibatches(self):
    """Generates a list of 2 minibatches to use in the tests.

    Returns:
      A list of 2 padded batches of examples.
      The structure is a list of the following:
      {
        'features': tf.tensor(float32) of shape(len, batch, dim)
        'paddings': tf.tensor(float32) of shape(len, batch)
      }
      where len=8, batch=2 and dim=3
    """

    # Specify the minibatch size
    max_seq_length = 8
    batch_size = 2
    feature_dims = 3

    # Specify the features and paddings for the first minibatch
    input_tensor1 = tf.reshape(
        tf.range(max_seq_length * batch_size * feature_dims, dtype=tf.float32),
        [max_seq_length, batch_size, feature_dims])
    paddings1 = tf.zeros(shape=[max_seq_length, batch_size], dtype=tf.float32)
    input_batch1 = py_utils.NestedMap(
        features=input_tensor1, paddings=paddings1)

    # Specify the features and paddings for the second minibatch
    # Also modify paddings2 to have a 1 at position [2, 0] in the tensor
    input_tensor2 = input_tensor1 + 1.0
    paddings2 = tf.zeros(shape=[max_seq_length, batch_size], dtype=tf.float32)
    paddings2 = tf.tensor_scatter_nd_update(paddings2, [[2, 0]], [1])
    input_batch2 = py_utils.NestedMap(
        features=input_tensor2, paddings=paddings2)

    # Create the list of minibatches
    input_batch_list = [input_batch1, input_batch2]

    return input_batch_list

  def _GetExpectedTestOutputs(self):
    """Returns the expected outputs for the tests.

    Returns:
      A 2D dictionary containing numpy (float32) arrays of the expected test
      outputs for PASS_THRU, MEAN and MEAN_STD while setting the
      use_weighted_frames setting to either True or False.
    """

    expected_outputs = {}
    expected_outputs['PASS_THRU'] = {
        False: np.array([[43.0, 44.0, 45.0], [46.0, 47.0, 48.0]])
    }
    expected_outputs['MEAN'] = {
        False: np.array(
          [[22.066668, 23.066668, 24.066668], [24.5, 25.5, 26.5]]),
        True: np.array(
          [[25.424778, 26.424774, 27.424776], [26.75111, 27.751108, 28.75111]])
    }
    expected_outputs['MEAN_STD'] = {
        False: np.array(
          [[22.066668, 23.066668, 24.066668, 14.026007, 14.026006, 14.026007],
           [24.5, 25.5, 26.5, 13.756816, 13.756816, 13.756816]]),
        True: np.array(
          [[25.424778, 26.424774, 27.424776, 12.595981, 12.595986, 12.595986],
           [26.75111, 27.751108, 28.75111, 12.927436, 12.927438, 12.927436]])
    }

    return expected_outputs

  def _TestHelperWithState(self, params, list_of_batches):
    """Returns the expected outputs for the tests.

    Args:
      params: Babelfish configuration parameters for setting up the
        cumulative_statistics_layer.
      list_of_batches: A list of padded batches of examples.
        The structure is a list of the following: {
        'features': tf.tensor(float32) of shape(len, batch, dim)
        'paddings': tf.tensor(float32) of shape(len, batch) }

    Returns:
      A dictionary containing numpy arrays of the expected test outputs.
      The structure is as follows:
      {
        'features': np.array(float32) of shape(len, batch, dim)
        'paddings': np.array(float32) of shape(len, batch)
      }
    """

    with self.session() as sess:
      tf.random.set_seed(_TF_RANDOM_SEED)
      network = params.Instantiate()

      batch_size = list_of_batches[0].features.shape[1]
      state = network.zero_state(network.theta, batch_size)

      for batch_t in list_of_batches:
        output = network.FProp(network.theta, batch_t, state)
        # Pass the output state over to the next batch as input state.
        state = output.state

      sess.run(
          tf.group(tf.global_variables_initializer(), tf.tables_initializer()))

      return sess.run(output)

  def _TestExpectedOutputs(self, stats_type, use_weighted_frames,
                           expected_last_frame_output):
    """Ochestrates the tests and checks that the expected outputs are correct.

    Args:
      stats_type: Specifies how the statistics will be accumulated. Options are
        'PASS_THRU', 'MEAN', and 'MEAN_STD'.
      use_weighted_frames: Whether or not to use frame-based weighting using a
        linear transform with sigmoid to determine the weight. Type is Boolean,
        True/False
      expected_last_frame_output: The last frame of output features once
        passed through the cumulative statistics layer. It is of the form:
          np.array(float32) with shape(batch, dim)
    """

    # Change basic params config to do the mean based statistics
    self.params.stats_type = stats_type
    self.params.use_weighted_frames = use_weighted_frames

    # Run the test
    output = self._TestHelperWithState(self.params, self.input_batch_list)

    # Check that for the 2nd minibatch that the input paddings are the same as
    # the output paddings.
    self.assertAllEqual(self.input_batch_list[1].paddings, output.paddings)

    # Check the size of the returned features. For the mean and standard
    # deviation case, the output should be double the dimensions.
    (max_seq_length, batch_size,
     feature_dims) = self.input_batch_list[0].features.shape.as_list()
    if stats_type == 'MEAN_STD':
      self.assertAllEqual((max_seq_length, batch_size, 2 * feature_dims),
                          output.features.shape)
    else:
      self.assertAllEqual((max_seq_length, batch_size, feature_dims),
                          output.features.shape)

    # Check that the last accumulated frame is correct
    last_frame = output.features[-1, :, :]
    self.assertAllClose(
        last_frame, expected_last_frame_output, rtol=1e-03, atol=1e-03)

  @parameterized.parameters(
      {
          'stats_type': 'PASS_THRU',
          'use_weighted_frames': False
      },
      {
          'stats_type': 'MEAN',
          'use_weighted_frames': False
      },
      {
          'stats_type': 'MEAN',
          'use_weighted_frames': True
      },
      {
          'stats_type': 'MEAN_STD',
          'use_weighted_frames': False
      },
      {
          'stats_type': 'MEAN_STD',
          'use_weighted_frames': True
      },
  )
  def testCumulativeStatisticsFProp(self, stats_type, use_weighted_frames):
    """Runs a test for a particular configuration.

    Args:
      stats_type: Specifies how the statistics will be accumulated. Options are
        'PASS_THRU', 'MEAN', and 'MEAN_STD'.
      use_weighted_frames: Whether or not to use frame-based weighting using a
        linear transform with sigmoid to determine the weight. Type is Boolean,
        True/False
    """

    # Run the particular test
    expected_last_frame_output = self.expected_outputs[stats_type][
        use_weighted_frames]
    self._TestExpectedOutputs(stats_type, use_weighted_frames,
                              expected_last_frame_output)


if __name__ == '__main__':
  tf.test.main()



"""Helper functions for modeling."""

from lingvo import compat as tf
from lingvo.core import py_utils


def GetLastSeqOutput(sequence_output, padding):
  """Gets the last sequence frame based on the paddings.

  Args:
    sequence_output: a tf.float32 tensor of size [time, batch, feature_dim]
    padding: a tf.float32 tensor of size [time, batch]

  Returns:
    last_frame: a tf.float32 tensor of size [batch, feature_dim]
  """
  sequence_output = py_utils.HasRank(sequence_output, 3)
  padding = py_utils.HasRank(padding, 2)

  sequence_output = py_utils.with_dependencies([
      py_utils.assert_equal(
          py_utils.GetShape(sequence_output, 2),
          py_utils.GetShape(padding, 2),
          message=('The first two dimensions of the two tensors '
                   f'must be the same: {py_utils.GetShape(sequence_output, 2)},'
                   f'{py_utils.GetShape(padding, 2)}')),
  ], sequence_output)

  # Last frame index
  idx_offset = 1
  # [batch]
  last_output_idx = tf.cast(
      tf.reduce_sum(1 - padding, axis=[0]) - idx_offset, dtype=tf.int32)

  batch_size = tf.shape(last_output_idx)[0]
  # [batch, 2]
  gather_idx = tf.stack(
      [last_output_idx, tf.range(batch_size, dtype=tf.int32)], axis=1)
  # [batch, feature_dim]
  last_frame = tf.gather_nd(sequence_output, gather_idx)
  return last_frame


def ComputeSimilaritySimple(logits, num_spks_per_batch, num_utts_per_spk):
  """Compute similarity score of each data to each speaker centroid.

  Each centroid is computed using all embeddings of a given speaker.

  Args:
    logits: output data of network Tensor([batch, dims])
    num_spks_per_batch: int, the number of speakers per batch
    num_utts_per_spk: int, the number of utterances per speaker

  Returns:
    cosine similarity matrix between each data and each speaker centroid.
    Tensor([batch, num_spks_per_batch])
  """
  # Normalize logits: [batch, dims]
  logits = tf.nn.l2_normalize(logits, axis=1)

  # logits_3d: [num_spks_per_batch, num_utts_per_spk, dims]
  logits_3d = tf.reshape(logits, [num_spks_per_batch, num_utts_per_spk, -1])

  # Compute centroid for each spk: [num_spks_per_batch, dims]
  centroid = tf.reduce_mean(logits_3d, axis=1)
  centroid = tf.nn.l2_normalize(centroid, axis=1)

  # Cosine similarity: [batch, num_spks_per_batch]
  cos_similarity = tf.matmul(logits, centroid, transpose_b=True)
  return cos_similarity


def ComputeSimilaritySplit(logits,
                           num_spks_per_batch,
                           num_utts_per_spk,
                           vary_number_of_enrollment_utterances_averaged=False):
  """Compute score of each data to each speaker centroid, with split batches.

  Each centroid is computed using all embeddings of a given speaker. The batch
  is split into the enrollment and test parts so that the loss is computed in
  a non-biased way.

  Args:
    logits: output data of network Tensor([batch, dims])
    num_spks_per_batch: int, the number of speakers per batch
    num_utts_per_spk: int, the number of utterances per speaker
    vary_number_of_enrollment_utterances_averaged: bool. If True, will
      randomize how many utterances are selected for enrollment. This makes the
      training challenge more difficult. If False (default), the average of all
      appropriate enrollment utterances is taken.

  Returns:
    cosine similarity matrix between each data and each speaker centroid.
    Tensor([batch, num_spks_per_batch])
  """
  batch_size = num_spks_per_batch * num_utts_per_spk

  # Normalize logits: [batch, dims]
  logits = tf.nn.l2_normalize(logits, axis=1)

  # logits_3d: [num_spks_per_batch, num_utts_per_spk, dims]
  logits_3d = tf.reshape(logits, [num_spks_per_batch, num_utts_per_spk, -1])

  def _GetCentroidAndData(part):
    """Compute Centroid and Data for one of two parts with length norm.

    This helper function extracts the speaker d-vectors and the speaker
    d-vector means for half of each speaker's data. When part is 0,
    all even numbered d-vector INDEXES are used for processing the d-vectors
    into test d-vectors and centroid models. When part is 1, all odd numbered
    d-vector INDEXES are utilized in the same manner.

    Args:
      part: Select which data part to use (0 or 1).

    Returns:
      centroid_part: For a data part, calculates the speaker specific mean.
      logits_part: The actual data part.
    """
    if part not in {0, 1}:
      raise ValueError('Must select an appropriate data part, either 0 or 1.')

    # logits_3d_part: [num_spks_per_batch, num_utts_per_spk // 2, dims]
    logits_3d_part = logits_3d[:, part::2, :]

    # logits_part: [batch // 2, dims]
    logits_part = tf.reshape(logits_3d_part, [batch_size // 2, -1])

    if vary_number_of_enrollment_utterances_averaged:
      # Compute centroid based on randomly selecting the number of utterances
      # to use followed by their random selection. The number of utterances used
      # is defined according to a uniform distribution from 1 up to the maximum
      # number of utterances.

      max_utt = num_utts_per_spk // 2

      # Randomly select the number of enrollment utterances to use for each
      # speaker and generate the sequence mask.
      lengths = tf.random.uniform(
          shape=(num_spks_per_batch,),
          minval=1,
          maxval=max_utt + 1,
          dtype=tf.dtypes.int32)
      mask = tf.sequence_mask(lengths, maxlen=max_utt, dtype=tf.dtypes.float32)

      # Randomize the sequence mask on each row.
      indices = tf.argsort(
          tf.random.uniform(shape=(num_spks_per_batch, max_utt)))
      mask_row_randomized = tf.gather(mask, indices, batch_dims=-1)

      # Calculate the average based on the mask
      sum_x = tf.reduce_sum(
          tf.math.multiply(logits_3d_part, mask_row_randomized[:, :,
                                                               tf.newaxis]),
          axis=1)
      sum_n = tf.cast(lengths, dtype=tf.float32)
      centroid_part = tf.math.divide(sum_x, sum_n[:, tf.newaxis])
    else:
      # Compute centroid for each spk: [num_spks_per_batch, dims]
      centroid_part = tf.reduce_mean(logits_3d_part, axis=1)

    centroid_part = tf.nn.l2_normalize(centroid_part, axis=1)
    return centroid_part, logits_part

  centroid_fst, logits_fst = _GetCentroidAndData(0)
  centroid_snd, logits_snd = _GetCentroidAndData(1)

  # cos_similarity_fst: [batch // 2, num_spks_per_batch]
  # cos_similarity_snd: [batch // 2, num_spks_per_batch]
  cos_similarity_fst = tf.matmul(logits_fst, centroid_snd, transpose_b=True)
  cos_similarity_snd = tf.matmul(logits_snd, centroid_fst, transpose_b=True)

  # cos_similarity:
  # [batch // 2, 2, num_spks_per_batch] -> [batch, num_spks_per_batch]
  cos_similarity = tf.reshape(
      tf.stack([cos_similarity_fst, cos_similarity_snd], axis=1),
      [batch_size, num_spks_per_batch])

  return cos_similarity
